const path = require("path");
const fs = require("fs");

const webpack = require("webpack");
const WebpackDevServer = require("webpack-dev-server");

const {
    choosePort,
    createCompiler,
    prepareProxy,
    prepareUrls
} = require("react-dev-utils/WebpackDevServerUtils");
const chalk = require("react-dev-utils/chalk");
const clearConsole = require("react-dev-utils/clearConsole");
const checkRequiredFiles = require("react-dev-utils/checkRequiredFiles");
const openBrowserTab = require("react-dev-utils/openBrowser");
// We require that you explicitly set browsers and do not fall back to browsers list defaults.
const { checkBrowsers } = require("react-dev-utils/browsersHelper");

module.exports = async options => {
    const { overrides, cwd } = options;
    const appIndexJs = overrides.entry || path.resolve(cwd, "src", "index.tsx");

    if (typeof options.openBrowser === "undefined") {
        options.openBrowser = true;
    }

    // Makes the script crash on unhandled rejections instead of silently
    // ignoring them. In the future, promise rejections that are not handled will
    // terminate the Node.js process with a non-zero exit code.
    process.on("unhandledRejection", err => {
        throw err;
    });

    const paths = require("./config/paths")({ appIndexJs, cwd });
    const configFactory = require("./config/webpack.config");
    const createDevServerConfig = require("./config/webpackDevServer.config");

    const useYarn = fs.existsSync(paths.yarnLockFile);
    const isInteractive = process.stdout.isTTY;

    // Warn and crash if required files are missing
    if (!checkRequiredFiles([paths.appHtml, paths.appIndexJs])) {
        process.exit(1);
    }

    let buildConfig = configFactory("development", { paths, options });

    if (typeof overrides.webpack === "function") {
        buildConfig = overrides.webpack(buildConfig);
    }

    try {
        await checkBrowsers(paths.appPath, isInteractive);

        const host = getHost();
        const port = await getPort(host);
        const https = process.env.HTTPS === "true";
        const protocol = https ? "https" : "http";
        const appName = require(paths.appPackageJson).name;
        const useTypeScript = fs.existsSync(paths.appTsConfig);
        const tscCompileOnError = process.env.TSC_COMPILE_ON_ERROR === "true";
        const urls = prepareUrls(protocol, host, port);

        // Create a webpack compiler that is configured with custom messages.
        const compiler = createCompiler({
            appName,
            config: buildConfig,
            urls,
            useYarn,
            useTypeScript,
            tscCompileOnError,
            webpack
        });
        // Load proxy config
        const proxySetting = require(paths.appPackageJson).proxy;
        const proxyConfig = prepareProxy(proxySetting, paths.appPublic);
        // Serve webpack assets generated by the compiler over a web server.
        const serverConfig = createDevServerConfig({
            host,
            port,
            https,
            allowedHost: urls.lanUrlForConfig,
            proxy: proxyConfig,
            paths
        });
        const devServer = new WebpackDevServer(serverConfig, compiler);

        // Launch WebpackDevServer.
        await devServer.start();

        if (isInteractive) {
            clearConsole();
        }

        console.log(chalk.cyan("Starting the development server...\n"));

        if (options.openBrowser) {
            // Powershell sets `SystemRoot`, but the `open` library is looking for `SYSTEMROOT`
            // so we need to make it happy by setting the right ENV variable.
            if (process.platform === "win32" && typeof process.env.SystemRoot !== "undefined") {
                process.env.SYSTEMROOT = process.env.SystemRoot;
            }
            openBrowserTab(urls.localUrlForBrowser);
        }

        ["SIGINT", "SIGTERM"].forEach(function (sig) {
            process.on(sig, async function () {
                await devServer.stop();
                process.exit();
            });
        });
    } catch (err) {
        if (err && err.message) {
            console.error(err);
        }
        process.exit(1);
    }
};

function getHost() {
    const host = process.env.HOST || "0.0.0.0";

    if (process.env.HOST) {
        console.log(
            chalk.cyan(
                `Attempting to bind to HOST environment variable: ${chalk.yellow(
                    chalk.bold(process.env.HOST)
                )}`
            )
        );
        console.log(
            `If this was unintentional, check that you haven't mistakenly set it in your shell.`
        );
        console.log(`Learn more here: ${chalk.yellow("https://bit.ly/CRA-advanced-config")}`);
        console.log();
    }

    return host;
}

async function getPort(host) {
    // Tools like Cloud9 rely on this.
    const defaultPort = parseInt(process.env.PORT, 10) || 3000;

    const port = await choosePort(host, defaultPort);
    if (port == null) {
        // We have not found a port.
        throw new Error(`No free port was found. Default port is ${defaultPort}`);
    }

    return port;
}
